@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2024 The pgagroal community
+ * Copyright (C) 2026 The pgagroal community
  *
  * Redistribution and use in source and binary forms, with or without modification,
  * are permitted provided that the following conditions are met:
@@ -28,45 +28,81 @@
 
 /* pgagroal */
 #include <pgagroal.h>
+#include <aes.h>
+#include <json.h>
 #include <logging.h>
+#include <management.h>
 #include <security.h>
+#include <stdint.h>
+#include <time.h>
 #include <utils.h>
+#include <utf8.h>
 
 /* system */
 #include <ctype.h>
+#include <err.h>
 #include <errno.h>
 #include <getopt.h>
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <unistd.h>
 #include <sys/mman.h>
 #include <sys/stat.h>
-#include <err.h>
+#include <unistd.h>
 
-#define DEFAULT_PASSWORD_LENGTH 64
-#define MIN_PASSWORD_LENGTH 8
+static int master_key(char* password, bool generate_pwd, int pwd_length, int32_t output_format);
+static int add_user(char* users_path, char* username, char* password, bool generate_pwd, int pwd_length, int32_t output_format);
+static int update_user(char* users_path, char* username, char* password, bool generate_pwd, int pwd_length, int32_t output_format);
+static int remove_user(char* users_path, char* username, int32_t output_format);
+static int list_users(char* users_path, int32_t output_format);
+static int create_response(char* users_path, struct json* json, struct json** response);
 
-#define ACTION_UNKNOWN     0
-#define ACTION_MASTER_KEY  1
-#define ACTION_ADD_USER    2
-#define ACTION_UPDATE_USER 3
-#define ACTION_REMOVE_USER 4
-#define ACTION_LIST_USERS  5
-
-static char CHARS[] = {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
-                       'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
-                       '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
-                       '!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '-', '_', '=', '+', '[', '{', ']', '}', '\\', '|', ';', ':',
-                       '\'', '\"', ',', '<', '.', '>', '/', '?'};
-
-static int master_key(char* password, bool generate_pwd, int pwd_length);
-static int add_user(char* users_path, char* username, char* password, bool generate_pwd, int pwd_length);
-static int update_user(char* users_path, char* username, char* password, bool generate_pwd, int pwd_length);
-static int remove_user(char* users_path, char* username);
-static int list_users(char* users_path);
-static char* generate_password(int pwd_length);
+// clang-format off
+const struct pgagroal_command command_table[] =
+{
+   {
+      .command = "master-key",
+      .subcommand = "",
+      .accepted_argument_count = {0},
+      .deprecated = false,
+      .action = MANAGEMENT_MASTER_KEY,
+      .log_message = "<master-key>",
+   },
+   {
+      .command = "user",
+      .subcommand = "add",
+      .accepted_argument_count = {0},
+      .deprecated = false,
+      .action = MANAGEMENT_ADD_USER,
+      .log_message = "<user add> [%s]",
+   },
+   {
+      .command = "user",
+      .subcommand = "edit",
+      .accepted_argument_count = {0},
+      .deprecated = false,
+      .action = MANAGEMENT_UPDATE_USER,
+      .log_message = "<user edit> [%s]",
+   },
+   {
+      .command = "user",
+      .subcommand = "del",
+      .accepted_argument_count = {0},
+      .deprecated = false,
+      .action = MANAGEMENT_REMOVE_USER,
+      .log_message = "<user del> [%s]",
+   },
+   {
+      .command = "user",
+      .subcommand = "ls",
+      .accepted_argument_count = {0},
+      .deprecated = false,
+      .action = MANAGEMENT_LIST_USERS,
+      .log_message = "<user ls>",
+   },
+};
+// clang-format on
 
 static void
 version(void)
@@ -87,12 +123,13 @@ usage(void)
    printf("\n");
    printf("Options:\n");
    printf("  -f, --file FILE         Set the path to a user file\n");
-   printf("                          Defaults to %s", PGAGROAL_DEFAULT_USERS_FILE);
+   printf("                          Defaults to %s\n", PGAGROAL_DEFAULT_USERS_FILE);
    printf("  -U, --user USER         Set the user name\n");
    printf("  -P, --password PASSWORD Set the password for the user\n");
    printf("  -g, --generate          Generate a password\n");
    printf("  -l, --length            Password length\n");
    printf("  -V, --version           Display version information\n");
+   printf("  -F, --format text|json  Set the output format\n");
    printf("  -?, --help              Display help\n");
    printf("\n");
    printf("Commands:\n");
@@ -117,10 +154,13 @@ main(int argc, char** argv)
    bool generate_pwd = false;
    int pwd_length = DEFAULT_PASSWORD_LENGTH;
    int option_index = 0;
-   int32_t action = ACTION_UNKNOWN;
+   size_t command_count = sizeof(command_table) / sizeof(struct pgagroal_command);
+   struct pgagroal_parsed_command parsed = {.cmd = NULL, .args = {0}};
+   int32_t output_format = MANAGEMENT_OUTPUT_FORMAT_TEXT;
 
    while (1)
    {
+      // clang-format off
       static struct option long_options[] =
       {
          {"user", required_argument, 0, 'U'},
@@ -129,10 +169,12 @@ main(int argc, char** argv)
          {"generate", no_argument, 0, 'g'},
          {"length", required_argument, 0, 'l'},
          {"version", no_argument, 0, 'V'},
+         {"format", required_argument, 0, 'F'},
          {"help", no_argument, 0, '?'}
       };
+      // clang-format on
 
-      c = getopt_long(argc, argv, "gV?f:U:P:l:",
+      c = getopt_long(argc, argv, "gV?f:U:P:l:F:",
                       long_options, &option_index);
 
       if (c == -1)
@@ -160,6 +202,21 @@ main(int argc, char** argv)
          case 'V':
             version();
             break;
+         case 'F':
+            if (!strncmp(optarg, "json", MISC_LENGTH))
+            {
+               output_format = MANAGEMENT_OUTPUT_FORMAT_JSON;
+            }
+            else if (!strncmp(optarg, "text", MISC_LENGTH))
+            {
+               output_format = MANAGEMENT_OUTPUT_FORMAT_TEXT;
+            }
+            else
+            {
+               warnx("pgagroal-cli: Format type is not correct");
+               exit(1);
+            }
+            break;
          case '?':
             usage();
             exit(1);
@@ -174,107 +231,92 @@ main(int argc, char** argv)
       errx(1, "Using the root account is not allowed");
    }
 
-   if (argc > 0)
+   if (!parse_command(argc, argv, optind, &parsed, command_table, command_count))
    {
-      if (!strcmp("master-key", argv[argc - 1]))
-      {
-         action = ACTION_MASTER_KEY;
-      }
-      else if (parse_command_simple(argc, argv, optind, "user", "add")
-               || parse_deprecated_command(argc, argv, optind, "add-user", NULL, "user add", 1, 6))
-      {
-         action = ACTION_ADD_USER;
-      }
-      else if (parse_command_simple(argc, argv, optind, "user", "edit")
-               || parse_deprecated_command(argc, argv, optind, "update-user", NULL, "user edit", 1, 6))
-      {
-         action = ACTION_UPDATE_USER;
-      }
-      else if (parse_command_simple(argc, argv, optind, "user", "del")
-               || parse_deprecated_command(argc, argv, optind, "remove-user", NULL, "user del", 1, 6))
-      {
-         action = ACTION_REMOVE_USER;
-      }
-      else if (parse_command_simple(argc, argv, optind, "user", "ls")
-               || parse_deprecated_command(argc, argv, optind, "list-users", NULL, "user ls", 1, 6))
-      {
-         action = ACTION_LIST_USERS;
-      }
+      usage();
+      goto error;
+   }
 
-      // exit immediatly if the action is not understood!
-      if (action == ACTION_UNKNOWN)
-      {
-         warnx("unknown command or subcommand <%s>", argv[optind]);
-         usage();
-         goto error;
-      }
+   // if here, the action is understood, but we need
+   // the file to operate onto!
+   // Therefore, if the user did not specify any config file
+   // the default one is used. Note that in the case of ACTION_MASTER_KEY
+   // there is no need for the file_path to be set, so setting to a default
+   // value does nothing.
+   // Setting the file also means we don't have to check against the file_path value.
+   if (file_path == NULL)
+   {
+      file_path = PGAGROAL_DEFAULT_USERS_FILE;
+   }
 
-      // if here, the action is understood, but we need
-      // the file to oeprate onto!
-      // Therefore, if the user did not specify any config file
-      // the default one is used. Note that in the case of ACTION_MASTER_KEY
-      // there is no need for the file_path to be set, so setting to a default
-      // value does nothing.
-      // Setting the file also means we don't have to check against the file_path value.
-      if (file_path == NULL)
+   if (parsed.cmd->action == MANAGEMENT_MASTER_KEY)
+   {
+      if (master_key(password, generate_pwd, pwd_length, output_format))
       {
-         file_path = PGAGROAL_DEFAULT_USERS_FILE;
+         errx(1, "Cannot generate master key");
       }
-
-      if (action == ACTION_MASTER_KEY)
+   }
+   else if (parsed.cmd->action == MANAGEMENT_ADD_USER)
+   {
+      if (add_user(file_path, username, password, generate_pwd, pwd_length, output_format))
       {
-         if (master_key(password, generate_pwd, pwd_length))
-         {
-            errx(1, "Error for master key");
-         }
+         errx(1, "Error for <user add>");
       }
-      else if (action == ACTION_ADD_USER)
+   }
+   else if (parsed.cmd->action == MANAGEMENT_UPDATE_USER)
+   {
+      if (update_user(file_path, username, password, generate_pwd, pwd_length, output_format))
       {
-         if (add_user(file_path, username, password, generate_pwd, pwd_length))
-         {
-            errx(1, "Error for <user add>");
-         }
+         errx(1, "Error for <user edit>");
       }
-      else if (action == ACTION_UPDATE_USER)
+   }
+   else if (parsed.cmd->action == MANAGEMENT_REMOVE_USER)
+   {
+      if (remove_user(file_path, username, output_format))
       {
-         if (update_user(file_path, username, password, generate_pwd, pwd_length))
-         {
-            errx(1, "Error for <user edit>");
-         }
+         errx(1, "Error for <user del>");
       }
-      else if (action == ACTION_REMOVE_USER)
+   }
+   else if (parsed.cmd->action == MANAGEMENT_LIST_USERS)
+   {
+      if (list_users(file_path, output_format))
       {
-
-         if (remove_user(file_path, username))
-         {
-            errx(1, "Error for <user del>");
-         }
-      }
-      else if (action == ACTION_LIST_USERS)
-      {
-
-         if (list_users(file_path))
-         {
-            errx(1, "Error for <user ls>");
-         }
-
+         errx(1, "Error for <user ls>");
       }
    }
 
    exit(0);
 
 error:
+
    exit(1);
 }
 
 static int
-master_key(char* password, bool generate_pwd, int pwd_length)
+master_key(char* password, bool generate_pwd, int pwd_length, int32_t output_format)
 {
    FILE* file = NULL;
    char buf[MISC_LENGTH];
    char* encoded = NULL;
+   size_t encoded_length;
    struct stat st = {0};
    bool do_free = true;
+   struct json* j = NULL;
+   struct json* outcome = NULL;
+   time_t start_t;
+   time_t end_t;
+
+   start_t = time(NULL);
+
+   if (pgagroal_management_create_header(MANAGEMENT_MASTER_KEY, 0, 0, output_format, &j))
+   {
+      goto error;
+   }
+
+   if (password != NULL)
+   {
+      do_free = false;
+   }
 
    if (pgagroal_get_home_directory() == NULL)
    {
@@ -307,7 +349,7 @@ master_key(char* password, bool generate_pwd, int pwd_length)
       }
       else
       {
-         warnx("Wrong permissions for ~/.pgagroal (must be 0700)");
+         warnx("Wrong permissions for directory <%s> (must be 0700)", &buf[0]);
          goto error;
       }
    }
@@ -315,6 +357,12 @@ master_key(char* password, bool generate_pwd, int pwd_length)
    memset(&buf, 0, sizeof(buf));
    snprintf(&buf[0], sizeof(buf), "%s/.pgagroal/master.key", pgagroal_get_home_directory());
 
+   if (pgagroal_exists(&buf[0]))
+   {
+      warnx("The file %s already exists, cannot continue", &buf[0]);
+      goto error;
+   }
+
    if (stat(&buf[0], &st) == -1)
    {
       /* Ok */
@@ -327,7 +375,7 @@ master_key(char* password, bool generate_pwd, int pwd_length)
       }
       else
       {
-         warnx("Wrong permissions for ~/.pgagroal/master.key (must be 0600)");
+         warnx("Wrong permissions for file <%s> (must be 0600)", &buf[0]);
          goto error;
       }
    }
@@ -339,28 +387,74 @@ master_key(char* password, bool generate_pwd, int pwd_length)
       goto error;
    }
 
+#if defined(HAVE_OSX)
+#define PGAGROAL_GETENV(name) getenv(name)
+#else
+#define PGAGROAL_GETENV(name) secure_getenv(name)
+#endif
+
    if (password == NULL)
    {
-      if (!generate_pwd)
+      if (generate_pwd)
       {
-         while (password == NULL)
+         if (pgagroal_generate_password(pwd_length, &password))
          {
-            printf("Master key (will not echo): ");
-            password = pgagroal_get_password();
-            printf("\n");
-
-            if (password != NULL && strlen(password) < MIN_PASSWORD_LENGTH)
-            {
-               printf("Invalid key length, must be at least %d chars.\n", MIN_PASSWORD_LENGTH);
-               free(password);
-               password = NULL;
-            }
+            do_free = false;
+            goto error;
          }
       }
       else
       {
-         password = generate_password(pwd_length);
-         do_free = false;
+         password = PGAGROAL_GETENV("PGAGROAL_PASSWORD");
+
+         if (password == NULL)
+         {
+            while (password == NULL)
+            {
+               printf("Master key (will not echo): ");
+               password = pgagroal_get_password();
+               printf("\n");
+
+               if (password != NULL)
+               {
+                  // Validate password is valid UTF-8
+                  if (!pgagroal_utf8_valid((const unsigned char*)password, strlen(password)))
+                  {
+                     printf("Invalid UTF-8 sequence in master key.\n");
+                     free(password);
+                     password = NULL;
+                     continue;
+                  }
+                  // Check character length
+                  size_t char_count = pgagroal_utf8_char_length((const unsigned char*)password, strlen(password));
+                  if (char_count == (size_t)-1)
+                  {
+                     printf("Error counting UTF-8 characters in master key.\n");
+                     free(password);
+                     password = NULL;
+                     continue;
+                  }
+                  if (char_count < MIN_PASSWORD_LENGTH)
+                  {
+                     printf("Invalid key length, must be at least %d characters.\n", MIN_PASSWORD_LENGTH);
+                     free(password);
+                     password = NULL;
+                     continue;
+                  }
+                  if (char_count > MAX_PASSWORD_CHARS)
+                  {
+                     printf("Master key too long (%zu characters). Maximum allowed: %d characters.\n", char_count, MAX_PASSWORD_CHARS);
+                     free(password);
+                     password = NULL;
+                     continue;
+                  }
+               }
+            }
+         }
+         else
+         {
+            do_free = false;
+         }
       }
    }
    else
@@ -368,16 +462,35 @@ master_key(char* password, bool generate_pwd, int pwd_length)
       do_free = false;
    }
 
-   pgagroal_base64_encode(password, strlen(password), &encoded);
+   end_t = time(NULL);
+
+   if (pgagroal_management_create_outcome_success(j, start_t, end_t, &outcome))
+   {
+      goto error;
+   }
+
+   if (output_format == MANAGEMENT_OUTPUT_FORMAT_JSON)
+   {
+      pgagroal_json_print(j, FORMAT_JSON);
+   }
+   else
+   {
+      pgagroal_json_print(j, FORMAT_TEXT);
+   }
+
+   pgagroal_base64_encode(password, strlen(password), &encoded, &encoded_length);
    fputs(encoded, file);
    free(encoded);
 
+   pgagroal_json_destroy(j);
+
    if (do_free)
    {
       free(password);
    }
 
    fclose(file);
+   file = NULL;
 
    chmod(&buf[0], S_IRUSR | S_IWUSR);
    printf("Master Key stored into %s\n", &buf[0]);
@@ -395,13 +508,27 @@ error:
    if (file)
    {
       fclose(file);
+      file = NULL;
    }
 
+   pgagroal_management_create_outcome_failure(j, 1, &outcome);
+
+   if (output_format == MANAGEMENT_OUTPUT_FORMAT_JSON)
+   {
+      pgagroal_json_print(j, FORMAT_JSON);
+   }
+   else
+   {
+      pgagroal_json_print(j, FORMAT_TEXT);
+   }
+
+   pgagroal_json_destroy(j);
+
    return 1;
 }
 
 static int
-add_user(char* users_path, char* username, char* password, bool generate_pwd, int pwd_length)
+add_user(char* users_path, char* username, char* password, bool generate_pwd, int pwd_length, int32_t output_format)
 {
    FILE* users_file = NULL;
    char line[MISC_LENGTH];
@@ -411,11 +538,24 @@ add_user(char* users_path, char* username, char* password, bool generate_pwd, in
    char* encrypted = NULL;
    int encrypted_length = 0;
    char* encoded = NULL;
+   size_t encoded_length;
    char un[MAX_USERNAME_LENGTH];
    int number_of_users = 0;
    bool do_verify = true;
    char* verify = NULL;
    bool do_free = true;
+   struct json* j = NULL;
+   struct json* outcome = NULL;
+   struct json* response = NULL;
+   time_t start_t;
+   time_t end_t;
+
+   start_t = time(NULL);
+
+   if (pgagroal_management_create_header(MANAGEMENT_ADD_USER, 0, 0, output_format, &j))
+   {
+      goto error;
+   }
 
    if (pgagroal_get_master_key(&master_key))
    {
@@ -457,6 +597,9 @@ username:
    }
 
    /* Verify */
+#ifdef HAVE_FREEBSD
+   fseek(users_file, 0, SEEK_SET);
+#endif
    while (fgets(line, sizeof(line), users_file))
    {
       ptr = strtok(line, ":");
@@ -481,31 +624,68 @@ username:
 password:
       if (generate_pwd)
       {
-         password = generate_password(pwd_length);
+         if (pgagroal_generate_password(pwd_length, &password))
+         {
+            do_free = false;
+            goto error;
+         }
          do_verify = false;
          printf("Password : %s", password);
+         do_free = true;
       }
       else
       {
-         printf("Password : ");
+         password = PGAGROAL_GETENV("PGAGROAL_PASSWORD");
 
-         if (password != NULL)
+         if (password == NULL)
          {
-            free(password);
-            password = NULL;
-         }
+            printf("Password : ");
 
-         password = pgagroal_get_password();
+            password = pgagroal_get_password();
+            do_free = true;
+            do_verify = true;
+         }
+         else
+         {
+            do_free = false;
+            do_verify = false;
+         }
       }
       printf("\n");
    }
 
-   for (int i = 0; i < strlen(password); i++)
+   // Validate password is valid UTF-8
+   if (!pgagroal_utf8_valid((const unsigned char*)password, strlen(password)))
    {
-      if ((unsigned char)(*(password + i)) & 0x80)
+      warnx("Invalid UTF-8 sequence in password");
+      if (do_free)
       {
-         goto password;
+         free(password);
       }
+      password = NULL;
+      goto password;
+   }
+   // Check character length
+   size_t char_count = pgagroal_utf8_char_length((const unsigned char*)password, strlen(password));
+   if (char_count == (size_t)-1)
+   {
+      warnx("Error counting UTF-8 characters in password");
+      if (do_free)
+      {
+         free(password);
+      }
+      password = NULL;
+      goto password;
+   }
+   if (char_count > MAX_PASSWORD_CHARS)
+   {
+      warnx("Password too long (%zu characters). Maximum allowed: %d characters.", char_count, MAX_PASSWORD_CHARS);
+      if (do_free)
+      {
+         free(password);
+      }
+      password = NULL;
+      goto password;
    }
 
    if (do_verify)
@@ -521,14 +701,61 @@ password:
       verify = pgagroal_get_password();
       printf("\n");
 
+      // Validate verification password is valid UTF-8
+      if (!pgagroal_utf8_valid((const unsigned char*)verify, strlen(verify)))
+      {
+         warnx("Invalid UTF-8 sequence in verification password. Please use valid UTF-8 encoding.");
+         free(verify);
+         verify = NULL;
+         if (do_free)
+         {
+            free(password);
+         }
+         password = NULL;
+         goto password;
+      }
+      // Check character count on verification password
+      size_t verify_char_count = pgagroal_utf8_char_length((const unsigned char*)verify, strlen(verify));
+      if (verify_char_count == (size_t)-1)
+      {
+         warnx("Error counting UTF-8 characters in verification password");
+         free(verify);
+         verify = NULL;
+         if (do_free)
+         {
+            free(password);
+         }
+         password = NULL;
+         goto password;
+      }
+      if (verify_char_count > MAX_PASSWORD_CHARS)
+      {
+         warnx("Verification password too long (%zu characters). Maximum allowed: %d characters.", verify_char_count, MAX_PASSWORD_CHARS);
+         free(verify);
+         verify = NULL;
+         if (do_free)
+         {
+            free(password);
+         }
+         password = NULL;
+         goto password;
+      }
       if (strlen(password) != strlen(verify) || memcmp(password, verify, strlen(password)) != 0)
       {
+         warnx("Passwords do not match");
+         if (do_free)
+         {
+            free(password);
+         }
+         password = NULL;
+         free(verify);
+         verify = NULL;
          goto password;
       }
    }
 
-   pgagroal_encrypt(password, master_key, &encrypted, &encrypted_length);
-   pgagroal_base64_encode(encrypted, encrypted_length, &encoded);
+   pgagroal_encrypt(password, master_key, &encrypted, &encrypted_length, ENCRYPTION_AES_256_CBC);
+   pgagroal_base64_encode(encrypted, encrypted_length, &encoded, &encoded_length);
 
    entry = pgagroal_append(entry, username);
    entry = pgagroal_append(entry, ":");
@@ -548,6 +775,30 @@ password:
    free(verify);
 
    fclose(users_file);
+   users_file = NULL;
+
+   end_t = time(NULL);
+
+   if (pgagroal_management_create_outcome_success(j, start_t, end_t, &outcome))
+   {
+      goto error;
+   }
+
+   if (create_response(users_path, j, &response))
+   {
+      goto error;
+   }
+
+   if (output_format == MANAGEMENT_OUTPUT_FORMAT_JSON)
+   {
+      pgagroal_json_print(j, FORMAT_JSON);
+   }
+   else
+   {
+      pgagroal_json_print(j, FORMAT_TEXT);
+   }
+
+   pgagroal_json_destroy(j);
 
    return 0;
 
@@ -566,13 +817,27 @@ error:
    if (users_file)
    {
       fclose(users_file);
+      users_file = NULL;
    }
 
+   pgagroal_management_create_outcome_failure(j, 1, &outcome);
+
+   if (output_format == MANAGEMENT_OUTPUT_FORMAT_JSON)
+   {
+      pgagroal_json_print(j, FORMAT_JSON);
+   }
+   else
+   {
+      pgagroal_json_print(j, FORMAT_TEXT);
+   }
+
+   pgagroal_json_destroy(j);
+
    return 1;
 }
 
 static int
-update_user(char* users_path, char* username, char* password, bool generate_pwd, int pwd_length)
+update_user(char* users_path, char* username, char* password, bool generate_pwd, int pwd_length, int32_t output_format)
 {
    FILE* users_file = NULL;
    FILE* users_file_tmp = NULL;
@@ -585,11 +850,24 @@ update_user(char* users_path, char* username, char* password, bool generate_pwd,
    char* encrypted = NULL;
    int encrypted_length = 0;
    char* encoded = NULL;
+   size_t encoded_length;
    char un[MAX_USERNAME_LENGTH];
    bool found = false;
    bool do_verify = true;
    char* verify = NULL;
    bool do_free = true;
+   struct json* j = NULL;
+   struct json* outcome = NULL;
+   struct json* response = NULL;
+   time_t start_t;
+   time_t end_t;
+
+   start_t = time(NULL);
+
+   if (pgagroal_management_create_header(MANAGEMENT_UPDATE_USER, 0, 0, output_format, &j))
+   {
+      goto error;
+   }
 
    memset(&tmpfilename, 0, sizeof(tmpfilename));
 
@@ -655,54 +933,135 @@ username:
 password:
             if (generate_pwd)
             {
-               password = generate_password(pwd_length);
+               if (pgagroal_generate_password(pwd_length, &password))
+               {
+                  do_free = false;
+                  goto error;
+               }
                do_verify = false;
                printf("Password : %s", password);
+               do_free = true;
             }
             else
             {
-               printf("Password : ");
+               password = PGAGROAL_GETENV("PGAGROAL_PASSWORD");
 
-               if (password != NULL)
+               if (password == NULL)
                {
-                  free(password);
-                  password = NULL;
-               }
+                  printf("Password : ");
 
-               password = pgagroal_get_password();
+                  password = pgagroal_get_password();
+                  do_free = true;
+                  do_verify = true;
+               }
+               else
+               {
+                  do_free = false;
+                  do_verify = false;
+               }
             }
             printf("\n");
          }
 
-         for (int i = 0; i < strlen(password); i++)
+         // Validate password is valid UTF-8
+         if (!pgagroal_utf8_valid((const unsigned char*)password, strlen(password)))
          {
-            if ((unsigned char)(*(password + i)) & 0x80)
+            warnx("Invalid UTF-8 sequence in password");
+            if (do_free)
             {
-               goto password;
+               free(password);
             }
+            password = NULL;
+            goto password;
+         }
+         // Check character length
+         size_t char_count = pgagroal_utf8_char_length((const unsigned char*)password, strlen(password));
+         if (char_count == (size_t)-1)
+         {
+            warnx("Error counting UTF-8 characters in password");
+            if (do_free)
+            {
+               free(password);
+            }
+            password = NULL;
+            goto password;
+         }
+         if (char_count > MAX_PASSWORD_CHARS)
+         {
+            warnx("Password too long (%zu characters). Maximum allowed: %d characters.", char_count, MAX_PASSWORD_CHARS);
+            if (do_free)
+            {
+               free(password);
+            }
+            password = NULL;
+            goto password;
          }
 
          if (do_verify)
          {
             printf("Verify   : ");
-
             if (verify != NULL)
             {
                free(verify);
                verify = NULL;
             }
-
             verify = pgagroal_get_password();
             printf("\n");
-
+            // Validate verification password is valid UTF-8
+            if (!pgagroal_utf8_valid((const unsigned char*)verify, strlen(verify)))
+            {
+               warnx("Invalid UTF-8 sequence in verification password. Please use valid UTF-8 encoding.");
+               free(verify);
+               verify = NULL;
+               if (do_free)
+               {
+                  free(password);
+               }
+               password = NULL;
+               goto password;
+            }
+            // Check character count on verification password
+            size_t verify_char_count = pgagroal_utf8_char_length((const unsigned char*)verify, strlen(verify));
+            if (verify_char_count == (size_t)-1)
+            {
+               warnx("Error counting UTF-8 characters in verification password");
+               free(verify);
+               verify = NULL;
+               if (do_free)
+               {
+                  free(password);
+               }
+               password = NULL;
+               goto password;
+            }
+            if (verify_char_count > MAX_PASSWORD_CHARS)
+            {
+               warnx("Verification password too long (%zu characters). Maximum allowed: %d characters.", verify_char_count, MAX_PASSWORD_CHARS);
+               free(verify);
+               verify = NULL;
+               if (do_free)
+               {
+                  free(password);
+               }
+               password = NULL;
+               goto password;
+            }
             if (strlen(password) != strlen(verify) || memcmp(password, verify, strlen(password)) != 0)
             {
+               warnx("Passwords do not match");
+               if (do_free)
+               {
+                  free(password);
+               }
+               password = NULL;
+               free(verify);
+               verify = NULL;
                goto password;
             }
          }
 
-         pgagroal_encrypt(password, master_key, &encrypted, &encrypted_length);
-         pgagroal_base64_encode(encrypted, encrypted_length, &encoded);
+         pgagroal_encrypt(password, master_key, &encrypted, &encrypted_length, ENCRYPTION_AES_256_CBC);
+         pgagroal_base64_encode(encrypted, encrypted_length, &encoded, &encoded_length);
 
          entry = NULL;
          entry = pgagroal_append(entry, username);
@@ -737,10 +1096,35 @@ password:
    free(verify);
 
    fclose(users_file);
+   users_file = NULL;
    fclose(users_file_tmp);
+   users_file_tmp = NULL;
 
    rename(tmpfilename, users_path);
 
+   end_t = time(NULL);
+
+   if (pgagroal_management_create_outcome_success(j, start_t, end_t, &outcome))
+   {
+      goto error;
+   }
+
+   if (create_response(users_path, j, &response))
+   {
+      goto error;
+   }
+
+   if (output_format == MANAGEMENT_OUTPUT_FORMAT_JSON)
+   {
+      pgagroal_json_print(j, FORMAT_JSON);
+   }
+   else
+   {
+      pgagroal_json_print(j, FORMAT_TEXT);
+   }
+
+   pgagroal_json_destroy(j);
+
    return 0;
 
 error:
@@ -757,11 +1141,13 @@ error:
    if (users_file)
    {
       fclose(users_file);
+      users_file = NULL;
    }
 
    if (users_file_tmp)
    {
       fclose(users_file_tmp);
+      users_file_tmp = NULL;
    }
 
    if (strlen(tmpfilename) > 0)
@@ -769,11 +1155,24 @@ error:
       remove(tmpfilename);
    }
 
+   pgagroal_management_create_outcome_failure(j, 1, &outcome);
+
+   if (output_format == MANAGEMENT_OUTPUT_FORMAT_JSON)
+   {
+      pgagroal_json_print(j, FORMAT_JSON);
+   }
+   else
+   {
+      pgagroal_json_print(j, FORMAT_TEXT);
+   }
+
+   pgagroal_json_destroy(j);
+
    return 1;
 }
 
 static int
-remove_user(char* users_path, char* username)
+remove_user(char* users_path, char* username, int32_t output_format)
 {
    FILE* users_file = NULL;
    FILE* users_file_tmp = NULL;
@@ -783,6 +1182,18 @@ remove_user(char* users_path, char* username)
    char* ptr = NULL;
    char un[MAX_USERNAME_LENGTH];
    bool found = false;
+   struct json* j = NULL;
+   struct json* outcome = NULL;
+   struct json* response = NULL;
+   time_t start_t;
+   time_t end_t;
+
+   start_t = time(NULL);
+
+   if (pgagroal_management_create_header(MANAGEMENT_REMOVE_USER, 0, 0, output_format, &j))
+   {
+      goto error;
+   }
 
    users_file = fopen(users_path, "r");
    if (!users_file)
@@ -844,10 +1255,35 @@ username:
    }
 
    fclose(users_file);
+   users_file = NULL;
    fclose(users_file_tmp);
+   users_file_tmp = NULL;
 
    rename(tmpfilename, users_path);
 
+   end_t = time(NULL);
+
+   if (pgagroal_management_create_outcome_success(j, start_t, end_t, &outcome))
+   {
+      goto error;
+   }
+
+   if (create_response(users_path, j, &response))
+   {
+      goto error;
+   }
+
+   if (output_format == MANAGEMENT_OUTPUT_FORMAT_JSON)
+   {
+      pgagroal_json_print(j, FORMAT_JSON);
+   }
+   else
+   {
+      pgagroal_json_print(j, FORMAT_TEXT);
+   }
+
+   pgagroal_json_destroy(j);
+
    return 0;
 
 error:
@@ -855,11 +1291,13 @@ error:
    if (users_file)
    {
       fclose(users_file);
+      users_file = NULL;
    }
 
    if (users_file_tmp)
    {
       fclose(users_file_tmp);
+      users_file_tmp = NULL;
    }
 
    if (strlen(tmpfilename) > 0)
@@ -867,15 +1305,40 @@ error:
       remove(tmpfilename);
    }
 
+   pgagroal_management_create_outcome_failure(j, 1, &outcome);
+
+   if (output_format == MANAGEMENT_OUTPUT_FORMAT_JSON)
+   {
+      pgagroal_json_print(j, FORMAT_JSON);
+   }
+   else
+   {
+      pgagroal_json_print(j, FORMAT_TEXT);
+   }
+
+   pgagroal_json_destroy(j);
+
    return 1;
 }
 
 static int
-list_users(char* users_path)
+list_users(char* users_path, int32_t output_format)
 {
    FILE* users_file = NULL;
    char line[MISC_LENGTH];
    char* ptr = NULL;
+   struct json* j = NULL;
+   struct json* outcome = NULL;
+   struct json* response = NULL;
+   time_t start_t;
+   time_t end_t;
+
+   start_t = time(NULL);
+
+   if (pgagroal_management_create_header(MANAGEMENT_LIST_USERS, 0, 0, output_format, &j))
+   {
+      goto error;
+   }
 
    users_file = fopen(users_path, "r");
    if (!users_file)
@@ -887,10 +1350,39 @@ list_users(char* users_path)
    while (fgets(line, sizeof(line), users_file))
    {
       ptr = strtok(line, ":");
+      if (strchr(ptr, '\n'))
+      {
+         continue;
+      }
       printf("%s\n", ptr);
    }
 
    fclose(users_file);
+   users_file = NULL;
+   users_file = NULL;
+
+   end_t = time(NULL);
+
+   if (pgagroal_management_create_outcome_success(j, start_t, end_t, &outcome))
+   {
+      goto error;
+   }
+
+   if (create_response(users_path, j, &response))
+   {
+      goto error;
+   }
+
+   if (output_format == MANAGEMENT_OUTPUT_FORMAT_JSON)
+   {
+      pgagroal_json_print(j, FORMAT_JSON);
+   }
+   else
+   {
+      pgagroal_json_print(j, FORMAT_TEXT);
+   }
+
+   pgagroal_json_destroy(j);
 
    return 0;
 
@@ -899,34 +1391,60 @@ error:
    if (users_file)
    {
       fclose(users_file);
+      users_file = NULL;
    }
 
    return 1;
 }
 
-static char*
-generate_password(int pwd_length)
+static int
+create_response(char* users_path, struct json* json, struct json** response)
 {
-   char* pwd;
-   size_t s;
-   time_t t;
+   struct json* r = NULL;
+   struct json* users = NULL;
+   FILE* users_file = NULL;
+   char line[MISC_LENGTH];
+   char* ptr = NULL;
 
-   s = pwd_length + 1;
+   *response = NULL;
 
-   pwd = calloc(1, s);
-   if (pwd == NULL)
+   if (pgagroal_json_create(&r))
    {
-      pgagroal_log_fatal("Couldn't allocate memory while generating password");
-      return NULL;
+      goto error;
    }
 
-   srand((unsigned)time(&t));
+   pgagroal_json_put(json, MANAGEMENT_CATEGORY_RESPONSE, (uintptr_t)r, ValueJSON);
 
-   for (int i = 0; i < s; i++)
+   if (pgagroal_json_create(&users))
    {
-      *((char*)(pwd + i)) = CHARS[rand() % sizeof(CHARS)];
+      goto error;
    }
-   *((char*)(pwd + pwd_length)) = '\0';
 
-   return pwd;
+   users_file = fopen(users_path, "r");
+   if (!users_file)
+   {
+      goto error;
+   }
+
+   while (fgets(line, sizeof(line), users_file))
+   {
+      ptr = strtok(line, ":");
+      if (strchr(ptr, '\n'))
+      {
+         continue;
+      }
+      pgagroal_json_append(users, (uintptr_t)ptr, ValueString);
+   }
+
+   pgagroal_json_put(r, "Users", (uintptr_t)users, ValueJSON);
+
+   *response = r;
+
+   return 0;
+
+error:
+
+   pgagroal_json_destroy(r);
+
+   return 1;
 }
