@@ -54,14 +54,6 @@
 #include <errno.h>
 #include <inttypes.h>
 
-#ifndef EVBACKEND_LINUXAIO
-#define EVBACKEND_LINUXAIO 0x00000040U
-#endif
-
-#ifndef EVBACKEND_IOURING
-#define EVBACKEND_IOURING 0x00000080U
-#endif
-
 extern char** environ;
 #if defined(HAVE_LINUX) || defined(HAVE_OSX)
 static bool env_changed = false;
@@ -1773,6 +1765,31 @@ hvsnprintf(char* buf, size_t n, const char* fmt, va_list ap)
          }
       }
 
+      /* Parse precision */
+      int precision = -1;
+      if (*p == '.')
+      {
+         p++;
+         if (*p == '*')
+         {
+            precision = va_arg(ap, int);
+            p++;
+         }
+         else
+         {
+            precision = 0;
+            while (isdigit((unsigned char)*p))
+            {
+               precision = precision * 10 + (*p - '0');
+               p++;
+            }
+         }
+         if (precision < 0)
+         {
+            precision = -1;
+         }
+      }
+
       /* Length modifier */
       enum { LM_NONE,
              LM_L,
@@ -1996,10 +2013,17 @@ int
 pgagroal_snprintf(char* buf, size_t n, const char* fmt, ...)
 {
    va_list ap;
+   va_list ap_copy;
    int ret;
 
    va_start(ap, fmt);
-   ret = hvsnprintf(buf, n, fmt, ap);
+   va_copy(ap_copy, ap);
+   ret = vsnprintf(buf, n, fmt, ap_copy);
+   va_end(ap_copy);
+   if (ret < 0)
+   {
+      ret = hvsnprintf(buf, n, fmt, ap);
+   }
    va_end(ap);
 
    return ret;
